from typing import TYPE_CHECKING, Any, Dict, Type, TypeVar

import attrs

from zenconfig.base import BaseConfig, Schema
from zenconfig.encoder import Encoder

AttrsInstance: Any
if TYPE_CHECKING:
    from attr import AttrsInstance
else:
    AttrsInstance = Any

C = TypeVar("C", bound=AttrsInstance)


class AttrsSchema(Schema[C]):
    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        return _load_nested(cls, cfg)

    def to_dict(self, config: C, encoder: Encoder) -> Dict[str, Any]:
        return encoder(attrs.asdict(config))


BaseConfig.register_schema(AttrsSchema(), attrs.has)


def _load_nested(cls: Type[C], cfg: Dict[str, Any]) -> C:
    """Load nested attrs."""
    kwargs: Dict[str, Any] = {}
    for field in attrs.fields(cls):
        if field.name not in cfg:
            continue
        value = cfg[field.name]
        if attrs.has(field.type):
            value = _load_nested(field.type, value)
        kwargs[field.name] = value
    return cls(**kwargs)



import contextlib
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, Type, TypeVar

from pydantic import BaseModel

from zenconfig.base import BaseConfig, Schema
from zenconfig.encoder import Encoder, encode

C = TypeVar("C", bound=BaseModel)


@dataclass
class PydanticSchema(Schema[C]):
    exclude_unset: bool = False
    exclude_defaults: bool = True

    def from_dict(self, cls: Type[C], cfg: Dict[str, Any]) -> C:
        return cls.parse_obj(cfg)

    def to_dict(self, config: C, encoder: Encoder) -> Dict[str, Any]:
        # Use pydantic encoders.
        return _encoder(config)(
            config.dict(
                exclude_unset=self.exclude_unset,
                exclude_defaults=self.exclude_defaults,
            )
        )


BaseConfig.register_schema(PydanticSchema(), lambda cls: issubclass(cls, BaseModel))


def _encoder(config: BaseModel) -> Encoder:
    def _enc(obj: Any) -> Any:
        with contextlib.suppress(TypeError):
            return config.__json_encoder__(obj)  # type: ignore
        return obj

    return partial(encode, _enc)



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

from zenconfig.base import BaseConfig, Format


@dataclass
class JSONFormat(Format):
    indent: int = 2
    sort_keys: bool = True
    ensure_ascii: bool = False

    def load(self, path: Path) -> Dict[str, Any]:
        return json.loads(path.read_text())

    def dump(
        self,
        path: Path,
        config: Dict[str, Any],
    ) -> None:
        path.write_text(
            json.dumps(
                config,
                indent=self.indent,
                sort_keys=self.sort_keys,
                ensure_ascii=self.ensure_ascii,
            ),
        )


BaseConfig.register_format(JSONFormat(), ".json")



import datetime
from collections import deque
from decimal import Decimal
from enum import Enum
from functools import partial
from ipaddress import (
    IPv4Address,
    IPv4Interface,
    IPv4Network,
    IPv6Address,
    IPv6Interface,
    IPv6Network,
)
from pathlib import Path
from re import Pattern
from typing import Any, Callable, Dict
from uuid import UUID

Encoder = Callable[[Any], Any]
Encoders = Dict[type, Encoder]
Data = Dict[str, Any]

BASE_ENCODERS: Encoders = {
    bytes: lambda o: o.decode(),
    datetime.date: lambda o: o.isoformat(),
    datetime.datetime: lambda o: o.isoformat(),
    datetime.time: lambda o: o.isoformat(),
    datetime.timedelta: lambda o: o.total_seconds(),
    Decimal: lambda o: int(o) if o.as_tuple().exponent >= 0 else float(o),
    Enum: lambda o: o.value,
    frozenset: list,
    deque: list,
    IPv4Address: str,
    IPv4Interface: str,
    IPv4Network: str,
    IPv6Address: str,
    IPv6Interface: str,
    IPv6Network: str,
    Path: str,
    Pattern: lambda o: o.pattern,
    set: list,
    tuple: list,
    UUID: str,
}


def _combine(encoders: Encoders, encoder: Encoder, obj: Any) -> Any:
    if encoders:
        for base in obj.__class__.__mro__[:-1]:
            if base not in encoders:
                continue
            return encoders[base](obj)
    return encoder(obj)


base_encoder: Encoder = partial(_combine, BASE_ENCODERS, lambda o: o)


def combine_encoders(encoders: Encoders, encoder: Encoder = base_encoder) -> Encoder:
    """Combine an encoder with additional encoders."""
    return partial(_combine, encoders, encoder)


def encode(encoder: Encoder, obj: Any) -> Any:
    """Recursively encode an object"""
    if isinstance(obj, dict):
        return {k: encode(encoder, v) for k, v in obj.items()}
    encoded = encoder(obj)
    if isinstance(encoded, list):
        return [encode(encoder, e) for e in encoded]
    return encoded



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zenconfig', 'zenconfig.formats', 'zenconfig.schemas']

package_data = \
{'': ['*']}

extras_require = \
{'attrs': ['attrs>=22,<23'],
 'pydantic': ['pydantic>=1,<2'],
 'toml': ['tomli>=2,<3', 'tomli-w>=1,<2'],
 'yaml': ['PyYAML>=6,<7']}

setup_kwargs = {
    'name': 'zenconfig',
    'version': '1.5.0',
    'description': 'Simple configuration loader for python.',
    'long_description': '# zen-config\n\n[![tests](https://github.com/gpajot/zen-config/workflows/Test/badge.svg?branch=main&event=push)](https://github.com/gpajot/zen-config/actions?query=workflow%3ATest+branch%3Amain+event%3Apush)\n[![version](https://img.shields.io/pypi/v/zenconfig?label=stable)](https://pypi.org/project/zenconfig/)\n[![python](https://img.shields.io/pypi/pyversions/zenconfig)](https://pypi.org/project/zenconfig/)\n\nSimple configuration loader for python.\n\nCompared to other solutions, the goal is to bring:\n- simple usage for simple use cases\n- multiple format support\n- use objects rather than plain dict to interact with the config\n- optionally use the power of pydantic for validation\n\n## Simple usage\nIf you don\'t want to configure much, pass the config path through the env variable `CONFIG`, and simply use:\n```python\nfrom dataclasses import dataclass\nfrom zenconfig import Config\n\n@dataclass\nclass MyConfig(Config):\n    some_key: str\n    some_optional_key: bool = False\n\n\ncfg = MyConfig(some_key="hello")\ncfg.save()\n...\ncfg = MyConfig.load()\ncfg.some_optional_key = True\ncfg.save()\n...\ncfg.clear()\n```\n\n## Config file loading\nWhen creating your config, you can specify at least one of those two attributes:\n- `ENV_PATH` the environment variable name containing the path to the config file, defaults to `CONFIG`\n- `PATH` directly the config path\n\n> 💡 When supplying both, if the env var is not set, it will use `PATH`.\n\nUser constructs will be expanded.\nIf the file does not exist it will be created.\nYou can specify the file mode via `Config.FILE_MODE`.\n\nThe config can be loaded from multiple files, see [fnmatch](https://docs.python.org/3/library/fnmatch.html) for syntax.\nNote that you will not be able to save if not handling exactly one file.\n\n## Read only\nIf you do not want to be able to modify the config from your code, you can use `ReadOnlyConfig`.\n\n## Supported formats\nCurrently, those formats are supported:\n- JSON\n- YAML - requires the `yaml` extra\n- TOML - requires the `toml` extra\n\nThe format is automatically inferred from the config file extension.\nWhen loading from multiple files, files can be of multiple formats.\n\nOther formats can be added by subclassing `Format`.\n\nTo register more formats: `Config.register_format(MyFormat(...), ".ext1", ".ext2")`.\n\n> 💡 You can re-register a format to change dumping options.\n\n## Supported schemas\nCurrently, those schemas are supported:\n- plain dict\n- dataclasses\n- pydantic models - requires the `pydantic` extra\n- attrs - requires the attrs extra\n\nThe schema is automatically inferred from the config class.\n\nOther schemas can be added by subclassing `Schema`.\n\nTo register more schemas: `Config.register_schema(MySchema(...), lambda cls: ...)`.\n\nYou can also force the schema by directly overriding the `SCHEMA` class attribute on your config.\nThis can be used to disable auto selection, or pass arguments to the schema instance.\n\n> ⚠️ When using pydantic, you have to supply the `ClassVar` type annotations\n> to all class variable you override\n> otherwise pydantic will treat those as its own fields and complain.\n\n### Conversions\nFor all schemas and formats, common built in types are handled [when dumping](https://github.com/gpajot/zen-config/blob/main/zenconfig/encoder.py).\n\n> ⚠️ Keep in mind that only `attrs` and `pydantic` support casting when loading the config.\n\nYou can add custom encoders with `Config.ENCODERS`. For `pydantic`, stick with [the standard way of doing it](https://pydantic-docs.helpmanual.io/usage/exporting_models/#json_encoders).\n\n## Contributing\nSee [contributing guide](https://github.com/gpajot/zen-config/blob/main/CONTRIBUTING.md).\n',
    'author': 'Gabriel Pajot',
    'author_email': 'gab@les-cactus.co',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gpajot/zen-config',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<3.12',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict

import tomli
import tomli_w

from zenconfig.base import BaseConfig, Format


@dataclass
class TOMLFormat(Format):
    multiline_strings: bool = True

    def load(self, path: Path) -> Dict[str, Any]:
        return tomli.loads(path.read_text())

    def dump(
        self,
        path: Path,
        config: Dict[str, Any],
    ) -> None:
        path.write_text(
            tomli_w.dumps(
                config,
                multiline_strings=self.multiline_strings,
            )
        )


BaseConfig.register_format(TOMLFormat(), ".toml")



import logging
from abc import ABC
from enum import IntEnum
from typing import Any, ClassVar, Dict, Type, TypeVar

from zenconfig.base import BaseConfig, ZenConfigError

logger = logging.getLogger(__name__)


class MergeStrategy(IntEnum):
    """Available merge strategies when handling multiple config files."""

    SHALLOW = 1
    DEEP = 2


C = TypeVar("C", bound="ReadOnlyConfig")


class ReadOnlyConfig(BaseConfig, ABC):
    """Abstract base class for supporting read only operations."""

    # Merge strategy chosen when handling multiple config files.
    MERGE_STRATEGY: ClassVar[MergeStrategy] = MergeStrategy.DEEP

    @classmethod
    def load(cls: Type[C]) -> C:
        """Load the configuration class from file(s)."""
        dict_config: Dict[str, Any] = {}
        for path in cls._paths():
            if not path.exists():
                continue
            fmt = cls._format(path)
            logger.debug(
                "using %s to load %s from %s",
                fmt.__class__.__name__,
                cls.__name__,
                path,
            )
            config = fmt.load(path)
            if not dict_config:
                dict_config = config
            elif cls.MERGE_STRATEGY is MergeStrategy.SHALLOW:
                dict_config.update(config)
            elif cls.MERGE_STRATEGY is MergeStrategy.DEEP:
                _deep_merge(dict_config, config)
            else:
                raise ZenConfigError(f"unsupported merge strategy {cls.MERGE_STRATEGY}")
        return cls._schema().from_dict(cls, dict_config)


def _deep_merge(a: Dict[str, Any], b: Dict[str, Any]) -> None:
    """Deep merge dictionaries."""
    for k, v in b.items():
        if k not in a or not isinstance(a[k], dict) or not isinstance(v, dict):
            a[k] = v
        else:
            _deep_merge(a[k], v)
